//	warpresize - Edge-directed resampling filter for VirtualDub
//	Copyright (C) 2005 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_VD1_VDFILTER_H
#define f_VD1_VDFILTER_H

#include <stdlib.h>
#include <stddef.h>
#include <new>

#include "filter.h"
#include "ScriptValue.h"

///////////////////////////////////////////////////////////////////////////
//
//	class VDFilter
//
//	This class handles most of the grimy work of creating the interface
//	between your filter and VirtualDub.

class VDFilterBase {
public:
	VDFilterBase(FilterActivation *fa, const FilterFunctions *ff);
	virtual ~VDFilterBase();

	// linkage routines

	virtual bool Init();
	virtual long GetParams()=0;
	virtual void Start();
	virtual bool Run() = 0;
	virtual void End();
	virtual bool Configure(HWND hwnd);
	virtual void GetSettingString(char *buf, int maxlen);
	virtual void GetScriptString(char *buf, int maxlen);
	virtual int Serialize(char *buf, int maxbuf) = 0;
	virtual int Deserialize(const char *buf, int maxbuf) = 0;

	static void __cdecl FilterDeinit   (FilterActivation *fa, const FilterFunctions *ff);
	static int  __cdecl FilterRun      (const FilterActivation *fa, const FilterFunctions *ff);
	static long __cdecl FilterParam    (FilterActivation *fa, const FilterFunctions *ff);
	static int  __cdecl FilterConfig   (FilterActivation *fa, const FilterFunctions *ff, HWND hWnd);
	static int  __cdecl FilterStart    (FilterActivation *fa, const FilterFunctions *ff);
	static int  __cdecl FilterEnd      (FilterActivation *fa, const FilterFunctions *ff);
	static bool __cdecl FilterScriptStr(FilterActivation *fa, const FilterFunctions *, char *, int);
	static void __cdecl FilterString2  (const FilterActivation *fa, const FilterFunctions *ff, char *buf, int maxlen);
	static int  __cdecl FilterSerialize    (FilterActivation *fa, const FilterFunctions *ff, char *buf, int maxbuf);
	static void __cdecl FilterDeserialize  (FilterActivation *fa, const FilterFunctions *ff, const char *buf, int maxbuf);

	enum { sScriptMethods = 0 };

protected:
	static INT_PTR CALLBACK StaticDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);
	virtual INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	bool DoConfigureDialog(HWND hwnd, LPCTSTR tmplName);

	// member variables
	FilterActivation *fa;
	const FilterFunctions *const ff;

	HWND	mhdlg;
	IFilterPreview *mifp;
};

///////////////////////////////////////////////////////////////////////////
//
//	class VDFilter
//
//	This is the class you should derive your filter from -- pass your class
//	as the template parameter.  It is critical that this be the most
//	derived name for your class, as it is the one used for placement new
//	and placement delete!
//
template<class T>
class VDFilter : public VDFilterBase {
public:
	VDFilter(FilterActivation *fa, const FilterFunctions *ff) : VDFilterBase(fa, ff) {}

	static int  __cdecl FilterInit     (FilterActivation *fa, const FilterFunctions *ff);
	static void __cdecl FilterCopy         (FilterActivation *fa, const FilterFunctions *ff, void *dst);
};

template<class T>
int  __cdecl VDFilter<T>::FilterInit     (FilterActivation *fa, const FilterFunctions *ff) {
	T *pThis = new(fa->filter_data) T(fa, ff);

	try {
		if (!pThis->Init()) {
			pThis->~T();
			return 1;
		}

		return 0;
	} catch(...) {
		pThis->~T();
		throw;
	}
}

template<class T>
void __cdecl VDFilter<T>::FilterCopy         (FilterActivation *fa, const FilterFunctions *ff, void *dst) {
	new(dst) T(*static_cast<T *>(reinterpret_cast<VDFilterBase *>(fa->filter_data)));
}

///////////////////////////////////////////////////////////////////////////
//
//	class VDFilterDefinition
//
//	This template creates the FilterDefinition structure for you based on
//	your filter class.
//
template<class T>
class VDFilterDefinition : public FilterDefinition {
public:
	VDFilterDefinition(const char *pszAuthor, const char *pszName, const char *pszDescription) {
		name			= pszName;
		desc			= pszDescription;
		maker			= pszAuthor;
		private_data	= NULL;
		inst_data_size	= sizeof(T);

		initProc		= T::FilterInit;
		deinitProc		= T::FilterDeinit;
		runProc			= T::FilterRun;
		paramProc		= T::FilterParam;
		configProc		= T::FilterConfig;
		stringProc		= NULL;
		startProc		= T::FilterStart;
		endProc			= T::FilterEnd;

		script_obj		= const_cast<CScriptObject *>(T::sScriptMethods ? static_cast<const CScriptObject *>(&sScriptObject) : (const CScriptObject *)0);
		fssProc			= T::FilterScriptStr;

		stringProc2		= T::FilterString2;
		serializeProc	= T::FilterSerialize;
		deserializeProc	= T::FilterDeserialize;
		copyProc		= T::FilterCopy;
	}

protected:
	static const CScriptObject sScriptObject;
};

template<class T>
const CScriptObject VDFilterDefinition<T>::sScriptObject={
	NULL, (ScriptFunctionDef *)T::sScriptMethods, NULL
};


///////////////////////////////////////////////////////////////////////////
//
//	utility functions
//

template<class T>
inline T *vdptrinc(T *ptr, ptrdiff_t offset) {
	return (T *)((char *)ptr + offset);
}

inline Pixel32 vdavgpixel(Pixel32 a, Pixel32 b) {
	return (a|b) - (((a^b)&0xfefefefe)>>1);
}

inline Pixel32 vdlerppixel(Pixel32 a, Pixel32 b, unsigned alpha) {
	return a + (((((b&0xff00ff) - (a&0xff00ff))*alpha + 0x00800080)>>8) & 0xff00ff) + (((((b&0xff00) - (a&0xff00))*alpha + 0x008000)>>8) & 0xff00);
}

// here is where we give compiler authors a heart attack
#define vditeratebitmap(bitmap)											\
	if (const int w = (bitmap).w)										\
		if (const int h = (bitmap).h)									\
			if(Pixel32 *dst = (bitmap).data + w)						\
				switch(long pitch = (bitmap).pitch) case 0: default:	\
					for(int y=h; y; --y, (dst = vdptrinc(dst, pitch)))	\
						for(int x=w; x; --x)							\
							switch(unsigned long& px = dst[x]) case 0: default:

#endif
